"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (g && (g = 0, op[0] && (_ = 0)), _) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
function streamie(handler, config) {
    var queue = {
        input: [],
        output: {
            success: [],
        },
    };
    var settings = {
        backpressureAt: config.backpressureAt || Infinity,
        concurrency: config.concurrency || 1,
        batchSize: config.batchSize || 1,
        maxBatchWait: config.maxBatchWait || Infinity,
        isFilter: config.isFilter || false,
        haltOnError: config.haltOnError || false,
    };
    var state = {
        count: {
            handling: 0,
        },
        lastHandledAt: null,
        get backpressure() {
            return queue.input.length >= settings.backpressureAt;
        },
        get isDrained() {
            return state.shouldDrain && (queue.input.length === 0) && (state.count.handling === 0) && (queue.output.success.length === 0);
        },
        isPaused: false,
        shouldDrain: false,
        isHalted: false,
    };
    var consumers = [];
    var eventHandlers = {
        onBackpressureRelease: new Set(),
        onDrained: new Set(),
        onError: new Set(),
    };
    var maxBatchWaitTimeout = null;
    // Internal functions
    function processInput() {
        return __awaiter(this, void 0, void 0, function () {
            var timeSinceLastHandled, startedWithBackpressure, itemsToHandle, handlerOutput, err_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (state.isPaused || state.isHalted || state.isDrained)
                            return [2 /*return*/];
                        if (queue.input.length === 0)
                            return [2 /*return*/];
                        timeSinceLastHandled = state.lastHandledAt && Date.now() - state.lastHandledAt;
                        // This top level condition establishes a normal condition under which we would not handle
                        // the items, as there aren't enough to justify a batch. However, we will handle them
                        // given the exceptions below
                        if (queue.input.length < settings.batchSize) {
                            if (
                            // If the queue is meant to be drained, even if the input queue is not a full batch,
                            // we will still handle it.
                            (!state.shouldDrain) &&
                                // If we have waited too long since the last handled batch, we will handle the items
                                (timeSinceLastHandled && (timeSinceLastHandled < settings.maxBatchWait))) {
                                // At this point, we're already not going to handle the items the process, but if
                                // there is a maxBatchWait configured, we will ensure that there is a timeout in place
                                // to call processInput after the maxBatchWait time has elapsed. This is because the
                                // qualification for maxBatchWait time could elapse and be qualified for a process, but
                                // no attempt to process would necessarily be invoked at that time.
                                if (settings.maxBatchWait) {
                                    if (maxBatchWaitTimeout)
                                        clearTimeout(maxBatchWaitTimeout);
                                    maxBatchWaitTimeout = setTimeout(function () {
                                        requestProcessInput();
                                    }, settings.maxBatchWait - timeSinceLastHandled);
                                }
                                return [2 /*return*/];
                            }
                        }
                        if (state.count.handling >= settings.concurrency)
                            return [2 /*return*/];
                        startedWithBackpressure = state.backpressure;
                        state.lastHandledAt = Date.now();
                        state.count.handling++;
                        itemsToHandle = queue.input.splice(0, settings.batchSize);
                        // const handlerInput = settings.batchSize === 1 ? itemsToHandle[0] : itemsToHandle;
                        if (startedWithBackpressure && !state.backpressure) {
                            eventHandlers.onBackpressureRelease.forEach(function (eventHandler) {
                                eventHandler();
                            });
                        }
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, handler(itemsToHandle)];
                    case 2:
                        handlerOutput = _a.sent();
                        // If the handler is a filter, we will only push to the output queue if the output is truthy.
                        // Otherwise, all outputs go to the output queue.
                        if (handlerOutput || !settings.isFilter)
                            queue.output.success.push({ input: itemsToHandle, output: handlerOutput });
                        return [3 /*break*/, 4];
                    case 3:
                        err_1 = _a.sent();
                        if (settings.haltOnError)
                            state.isHalted = true;
                        eventHandlers.onError.forEach(function (eventHandler) {
                            eventHandler({ input: itemsToHandle, error: err_1 });
                        });
                        return [3 /*break*/, 4];
                    case 4:
                        state.count.handling--;
                        return [4 /*yield*/, requestProcessOutput()];
                    case 5:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    }
    function processOutput() {
        if (state.isPaused || state.isHalted || state.isDrained)
            return;
        // TODO should allow different strategies, but for now we will say that if any consumer
        // is backpressured, no other consumers will be pushed to, as this could allow a queue
        // to grow indefinitely.
        if (consumers.some(function (consumer) { return consumer.streamie.state.backpressure; }))
            return;
        var success = queue.output.success.shift();
        if (success) {
            consumers.forEach(function (consumer) {
                consumer.push(success);
            });
        }
    }
    function requestProcessInput() {
        return __awaiter(this, void 0, void 0, function () {
            var isDrained;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        isDrained = state.isDrained;
                        // Allow this to be more complicated logic in future.
                        return [4 /*yield*/, processInput()];
                    case 1:
                        // Allow this to be more complicated logic in future.
                        _a.sent();
                        if (!isDrained && state.isDrained) {
                            eventHandlers.onDrained.forEach(function (eventHandler) {
                                eventHandler();
                            });
                        }
                        return [2 /*return*/];
                }
            });
        });
    }
    function requestProcessOutput() {
        return __awaiter(this, void 0, void 0, function () {
            var isDrained;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        isDrained = state.isDrained;
                        // Allow this to be more complicated logic in future.
                        return [4 /*yield*/, processOutput()];
                    case 1:
                        // Allow this to be more complicated logic in future.
                        _a.sent();
                        if (!isDrained && state.isDrained) {
                            eventHandlers.onDrained.forEach(function (eventHandler) {
                                eventHandler();
                            });
                        }
                        return [2 /*return*/];
                }
            });
        });
    }
    // Public functions
    function push() {
        var _a;
        var items = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            items[_i] = arguments[_i];
        }
        if (state.isHalted)
            throw new Error('Cannot push to a halted streamie.');
        if (state.shouldDrain)
            throw new Error("Cannot push to a ".concat(state.isDrained ? 'drained' : 'draining', " streamie."));
        (_a = queue.input).push.apply(_a, items);
        requestProcessInput();
    }
    function map(handler, config) {
        var nextStreamie = streamie(handler, config);
        nextStreamie.onBackpressureRelease(function () { return requestProcessOutput(); });
        consumers.push({
            push: function (item) {
                nextStreamie.push(item.output);
            },
            streamie: nextStreamie,
        });
        return nextStreamie;
    }
    function filter(handler, config) {
        return map(handler, __assign(__assign({}, config), { isFilter: true }));
    }
    // TODO add reduce, flatMap, etc.
    function pause(shouldPause) {
        state.isPaused = shouldPause !== null && shouldPause !== void 0 ? shouldPause : !state.isPaused;
        if (!state.isPaused)
            requestProcessInput();
    }
    function drain() {
        state.shouldDrain = true;
        requestProcessInput();
    }
    function onBackpressureRelease(eventHandler) {
        eventHandlers.onBackpressureRelease.add(eventHandler);
    }
    function onDrained(eventHandler) {
        eventHandlers.onDrained.add(eventHandler);
    }
    function onError(eventHandler) {
        eventHandlers.onError.add(eventHandler);
    }
    return {
        push: push,
        map: map,
        filter: filter,
        pause: pause,
        drain: drain,
        state: {
            get backpressure() {
                return state.backpressure;
            },
            get count() {
                return __assign({}, state.count);
            },
            get isPaused() {
                return state.isPaused;
            },
            get isDrained() {
                return state.isDrained;
            },
            get isHalted() {
                return state.isHalted;
            },
        },
        onBackpressureRelease: onBackpressureRelease,
        onDrained: onDrained,
        onError: onError,
        promise: new Promise(function (resolve, reject) {
            onDrained(function () { return resolve(null); });
            if (settings.haltOnError)
                onError(function (_a) {
                    var error = _a.error;
                    return reject(error);
                });
        }),
    };
}
exports.default = streamie;
