export type Handler<HI, HO> = (input: HI[]) => HO;
export type Config = {
    backpressureAt?: number;
    concurrency?: number;
    batchSize?: number;
    maxBatchWait?: number;
    isFilter?: boolean;
    haltOnError?: boolean;
};
export type Streamie<HI extends any, HO extends any> = {
    push: (item: HI) => void;
    map: <NHO extends any, C extends Config>(handler: Handler<HO, NHO>, config: C) => Streamie<HO, NHO>;
    filter: <NHO extends any, C extends Omit<Config, 'isFilter'>>(handler: Handler<HO, NHO>, config: C) => Streamie<HO, NHO>;
    pause: (shouldPause?: boolean) => void;
    drain: () => void;
    onBackpressureRelease: (eventHandler: () => void) => void;
    onDrained: (eventHandler: () => void) => void;
    onError: (eventHandler: (payload: {
        input: HI[];
        error: any;
    }) => void) => void;
    state: {
        backpressure: boolean;
        isPaused: boolean;
        isDrained: boolean;
        isHalted: boolean;
        count: {
            handling: number;
        };
    };
    promise: Promise<null>;
};
export type Consumer<HI extends any, HO extends any, NHO extends any> = {
    push: (item: {
        input: HI[];
        output: HO;
    }) => void;
    streamie: Streamie<HO, NHO>;
};
